package CH.psi.midas;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.math.*;
import java.beans.*;
import java.util.*;
import javax.swing.*;

public class MJHist extends javax.swing.JPanel implements Serializable,ActionListener
{

  public MJHist()
  {
    //{{INIT_CONTROLS
	setLayout(null);
	setSize(72,24);
	//}}
  }
  
  public MJHist(String equipment)
  {
    this();
    this.equipment = equipment;
  }

  //{{DECLARE_CONTROLS
  //}}

  public String getEquipment()
  {
    return this.equipment;
  }

  public void setEquipment(String e)
  {
    this.equipment = e;
  }

  public OdbKey getOdbKey()
  {
    return this.odbKey;
  }
  
  public boolean isDrawLabels() {
    return this.drawLabels;
  }
  
  public void connect(Experiment exp, OdbKey odbKey) throws 
                IOException, TooManyListenersException, OdbKeyNotFoundException 
  {
    this.experiment = exp;
    this.drawLabels = true;
    this.odbKey = odbKey;
    
    // initialize popup menu
    buildPopup();
    
    // retrieve current time from server
    String r = experiment.rpcCall("cm_time");
    int end = Integer.parseInt(experiment.replyParam(r,1));

    // remember time offset for later
    timeOffset = end - (int) (new Date().getTime()/1000);
    
    // go one hour back
    int start = end - 3600;
    
    // Initialize scales
    scaleXMax = 0;
    scaleXMin = -1;  // 1h ago
    scaleYMin = 0;
    scaleYMax = 0;

    // set history directory
    String path = experiment.getOdbValue("/Logger/Data dir");
    experiment.rpcCall("hs_set_path&"+path);

    // initialize data curves
    dataCurve = new Vector();
    for (int i=0 ; i<odbKey.getNumValues() ; i++) {
    
      // extract event and variables names
      String str = odbKey.toString(i);
      int index =  str.indexOf('/');
      if (index > 0) {
        String eventName = str.substring(0, index-1);
        String varName = str.substring(index+2, str.length());
        
        dataCurve.addElement(new DataCurve(this, eventName, varName, colors[i%8], start, end));
      }
    }

    // Register mouse listener
    this.addMouseListener(new MouseAdapter() {
      public void mouseClicked(MouseEvent e) {
        if (e.getModifiers() == InputEvent.BUTTON3_MASK) {
          doPopup(e.getPoint());
        }
      }
    });

    // start update timer
    timer = new Timer(10000, this);
    timer.start();
  }
  
  public void disconnect() {
    // unlink curves to close ODB hotlinks
    try {
      for (int i=0 ; i<dataCurve.size() ; i++) {
        DataCurve dc = (DataCurve)dataCurve.elementAt(i);
        dc.unlink();
      }
    } catch (IOException e) { System.out.println(e); }
  }

  public void actionPerformed(ActionEvent e) {
    String cmd = e.getActionCommand();
    
    if (e.getSource() instanceof Timer)
      // do repaint every 10 sec
      this.repaint();

    else if (cmd.equals("Variables...")) {

      this.setCursor(new Cursor(Cursor.WAIT_CURSOR));
      SwingUtilities.invokeLater(new Runnable() {
        public void run () {
          doVarDialog();
        }
      });
    }
  
    else if (cmd.equals("Draw labels")) {
      this.drawLabels = ((JCheckBoxMenuItem)e.getSource()).getState();
      this.repaint();
    }
    else if (cmd.equals("Start Y at zero")) {
      this.startYZero = ((JCheckBoxMenuItem)e.getSource()).getState();
      this.repaint();
    }

    else if (e.getSource() instanceof JMenuItem) {
      if (e.getActionCommand().equals("1m"))
        this.scaleXMin = -1/60.0;
      else if (cmd.equals("10m"))
        this.scaleXMin = -10/60.0;
      else if (cmd.equals("1h"))
        this.scaleXMin = -1;
      else if (cmd.equals("6h"))
        this.scaleXMin = -6;
      else if (cmd.equals("24h"))
        this.scaleXMin = -24;
      else if (cmd.equals("7d"))
        this.scaleXMin = -7*24;
        
      // recreate curves
      this.setCursor(new Cursor(Cursor.WAIT_CURSOR));
      this.getToolkit().sync();
      SwingUtilities.invokeLater(new Runnable() {
        public void run () {
          recreateCurves();
        }
      });
    }
  }

  private void recreateCurves() {
    try {
      String r = experiment.rpcCall("cm_time");
      int end = Integer.parseInt(experiment.replyParam(r,1));
      int start = end + (int)(scaleXMin*3600);
      
      for (int i=0 ; i<dataCurve.size() ; i++) {
        DataCurve dc = (DataCurve)dataCurve.elementAt(i);
        dc.unlink();
      }
        
      for (int i=0 ; i<dataCurve.size() ; i++) {
        DataCurve dc = (DataCurve)dataCurve.elementAt(i);
        DataCurve ndc = new DataCurve(this, dc.getEventName(), dc.getVarName(), 
                                            dc.getColor(), start, end);
                                              
        dataCurve.setElementAt(ndc, i);                                    
      }
                                              
    } catch (Exception ex) { System.out.println(ex); }
      finally { 
        this.setCursor(Cursor.getDefaultCursor()); 
      }

  }

  private void buildPopup() {
    popup = new JPopupMenu();
    JMenuItem  mi;
    
    JMenu sm = new JMenu("Time Scale");
    mi = sm.add(new JMenuItem("1m"));
    mi.addActionListener(this);
    mi = sm.add(new JMenuItem("10m"));
    mi.addActionListener(this);
    mi = sm.add(new JMenuItem("1h"));
    mi.addActionListener(this);
    mi = sm.add(new JMenuItem("6h"));
    mi.addActionListener(this);
    mi = sm.add(new JMenuItem("24h"));
    mi.addActionListener(this);
    mi = sm.add(new JMenuItem("7d"));
    mi.addActionListener(this);
    popup.add(sm);
    
    mi = popup.add("Variables...");
    mi.addActionListener(this);
    
    JCheckBoxMenuItem cmi = new JCheckBoxMenuItem("Draw labels", this.drawLabels);
    cmi.addActionListener(this);
    popup.add(cmi);
    
    cmi = new JCheckBoxMenuItem("Start Y at zero", this.startYZero);
    cmi.addActionListener(this);
    popup.add(cmi);
  }

  private void doPopup(Point p) {
    popup.show(this, p.x, p.y);
  }

  private void doVarDialog() {
    Vector eventNames = new Vector();
    Vector eventIDs = new Vector();
    Vector eventVarNames = new Vector();
    Vector varSelection = new Vector();
    
    try {
      // Enumerate events
      String r = experiment.rpcCall("hs_enum_events&0&32000&32000");
    
      if (experiment.replyParam(r, 0).equals("1")) {
        
        int n = Integer.parseInt(experiment.replyParam(r, 1));
        for (int i=0 ; i<n ; i++) {
          eventNames.addElement(experiment.replyParam(r, 2+i));
          eventIDs.addElement(experiment.replyParam(r, 4+n+i));
        }
        
        // Enumerate variables
        for (int i=0 ; i<n ; i++) {
          
          r = experiment.rpcCall("hs_enum_vars&0&"+eventIDs.elementAt(i)+"&64000");

          int ib, ie;
          Vector varNames = new Vector();

          // skip status
          for (ib = 0 ; r.charAt(ib) != '&' ; ib++);
          ib++;

          // get n
          for (ie = ib+1; r.charAt(ie) != '&' ; ie++);
          int nVars = Integer.parseInt(r.substring(ib, ie));
          ib = ie+1;

          // extract names
          for (int j=0 ; j<nVars ; j++) {
            for (ie = ib+1; r.charAt(ie) != '&' ; ie++);
            varNames.addElement(r.substring(ib, ie));
            ib = ie+1;
          }
          
          eventVarNames.addElement(varNames);
        }
   
        // fill varSelection with current curves
        for (int i=0 ; i<dataCurve.size() ; i++) {
          DataCurve dc = (DataCurve)dataCurve.elementAt(i);
          varSelection.addElement(dc.getEventName()+" / "+dc.getVarName());
        }

        // remove wait cursor
        this.setCursor(Cursor.getDefaultCursor());
        
        // create dialog box
        MJHistVarDialog d = new MJHistVarDialog(eventNames, eventVarNames, varSelection);
        d.show();
      
        if (d.getReturnValue() == JOptionPane.CLOSED_OPTION) {
        
          // Remove old curves
            for (int i=0 ; i<dataCurve.size() ; i++) {
              DataCurve dc = (DataCurve)dataCurve.elementAt(i);
              dc.unlink();
            }
          dataCurve.removeAllElements();
        
          for (int i=0 ; i<varSelection.size() ; i++) {
          
            // extract event and variables names
            String var = (String)varSelection.elementAt(i);
            int index =  var.indexOf('/');
            String eventName = var.substring(0, index-1);
            String varName = var.substring(index+2, var.length());

            // get scale
            r = experiment.rpcCall("cm_time");
            int end = Integer.parseInt(experiment.replyParam(r,1));
            int start = end + (int)(scaleXMin*3600);

            // create curve
            dataCurve.addElement(new DataCurve(this, eventName, varName, colors[i%8], start, end));
            
            // save curve in ODB
            odbKey.setTextAt(var, i, 64);
          }
          
          // truncate ODB key
          odbKey.setNumValues(varSelection.size());
        }
      }
    } catch (Exception e) { System.out.println(e); }
    
  }
  
  static public void main(String args[])
  {
    class DriverFrame extends java.awt.Frame {
	  public DriverFrame() {
	    addWindowListener(new java.awt.event.WindowAdapter() {
          public void windowClosing(java.awt.event.WindowEvent event)
		    {
			dispose();	  // free the system resources
			System.exit(0); // close the application
			}
		});
		
		this.setLayout(null);
		this.setSize(72,24);
		this.add(new MJLabel());
	  }
	}

    new DriverFrame().show();
  }

  public void paint(Graphics g) {
    
    Rectangle r = this.getBounds();
    Rectangle rc = g.getClipBounds();
    
    r.x = r.y = 0;
    
    if (!r.equals(rc)) {
      this.repaint();
      return;
    }

    // draw frame 
    g.setColor(Color.white);
    g.fillRect(r.x, r.y, r.width-1, r.height-1);
    g.setColor(Color.black);
    g.drawRect(r.x, r.y, r.width-1, r.height-1);

    // retrieve individual axis limits
    scaleYMin = scaleYMax = 0;
    for (int i=0 ; i<dataCurve.size() ; i++) {
      double min = ((DataCurve)dataCurve.elementAt(i)).getScaleYMin();
      double max = ((DataCurve)dataCurve.elementAt(i)).getScaleYMax();
      if (!startYZero && i==0)
        scaleYMin = min;
      if (min < scaleYMin)
        scaleYMin = min;
      if (max > scaleYMax)
        scaleYMax = max;
    }
    
    // calculcate global limits
    double range = scaleYMax - scaleYMin;
    if (range == 0)
      range = 1;
    double ymin = scaleYMin - 0.1*range;
    double ymax = scaleYMax + 0.1*range;
    
    // draw axis
    g.setFont(new Font("SansSerif", Font.PLAIN, 10));
    int labelHeight = g.getFontMetrics().getHeight();
    int labelWidth = drawVAxis(g, 0, 0, r.height, 0, ymin, ymax)+5;
    
    int x = labelWidth;
    int y = r.height-labelHeight;
    int width = r.width-labelWidth-5;
    int height = r.height-labelHeight-5;
    
    // draw axis frame
    drawHAxis(g, x, y, width, height, scaleXMin, scaleXMax);
    drawVAxis(g, x, y, height, width, ymin, ymax);
    g.drawLine(x, y-height, x+width, y-height);
    g.drawLine(x+width, y, x+width, y-height);
    
    // draw curves
    for (int i=0 ; i<dataCurve.size() ; i++)
      ((DataCurve)dataCurve.elementAt(i)).paint(g, scaleXMin, scaleXMax, ymin, ymax, 
                                                x, y, width, height, timeOffset);
                  
    // draw labels
    if (this.drawLabels) 
      for (int i=0; i<dataCurve.size() ; i++) {
        DataCurve dc = (DataCurve)dataCurve.elementAt(i);
        int strWidth = g.getFontMetrics().stringWidth(dc.getVarName());
        int strHeight = g.getFontMetrics().getHeight();

        g.setColor(Color.white);
        g.fillRect(x+5, y-height+5+i*(strHeight+5), strWidth+6, strHeight);
        g.setColor(Color.black);
        g.drawString(dc.getVarName(), x+8, y-height+5+i*(strHeight+5)+g.getFontMetrics().getAscent());
        g.setColor(dc.getColor());
        g.drawRect(x+5, y-height+5+i*(strHeight+5), strWidth+6, strHeight);
    }
    
  } 
 
  private static final double LOG2 = Math.log(2)/Math.log(10);
  private static final double LOG5 = Math.log(5)/Math.log(10);
  private static final double LN10 = Math.log(10);

  private void drawHAxis(Graphics g, int x1, int y1, int width, int tickGrid, 
                          double xmin, double xmax) {

    double labelDx, majorDx, maxWidth;
    int    xScreen, strWidth;
    String str;
    double base[] = {1,2,5,10,20,50,100,200,500,1000};

    int minTickDist = 5;
    int tickLabel   = 10;
    int tickMajor   = 10;
    int tickMinor   = 5;
    int distLabel   = -g.getFontMetrics().getAscent();

    int maxTick = (int) ((double) (width/minTickDist) + 1);

    double dx = (xmax - xmin)/ (double) (width/minTickDist);

    double tmp = Math.log(dx) / LN10;
    int intDx = (int) tmp;
    double fracDx = tmp - intDx;
    
    if (fracDx <0) { 
      fracDx +=1; 
      intDx -=1; 
    }

    int tickBase = fracDx < LOG2 ? 1 :
                   fracDx < LOG5 ? 2 : 3;
    int majorBase = tickBase + 1;
    int labelBase = tickBase + 1;

    // rounding up of dx, labelDx
    dx = Math.pow(10, intDx) * base[tickBase];
    majorDx = Math.pow(10, intDx) * base[majorBase];
    labelDx = majorDx;

    // number of significant digits
    int nSig1 = 0;
    if (xmin > 0)
      nSig1 = (int) Math.floor(Math.log(Math.abs(xmin))/LN10) -
              (int) Math.floor(Math.log(Math.abs(labelDx))/LN10) + 1;

    int nSig2 = 0;
    if (xmax > 0)
      nSig2 = (int) Math.floor(Math.log(Math.abs(xmax))/LN10) -
              (int) Math.floor(Math.log(Math.abs(labelDx))/LN10) + 1;

    nSig1 = Math.max(nSig1, nSig2);
    nSig1 = Math.max(nSig1, 4);

    // determination of maximal width of labels 
    str = expFormat(Math.floor(xmin/dx)*dx, nSig1);
    maxWidth = g.getFontMetrics().stringWidth(str);

    str = expFormat(Math.floor(xmax/dx)*dx, nSig1);
    maxWidth = Math.max(maxWidth, g.getFontMetrics().stringWidth(str));

    str = expFormat(Math.floor(xmin/dx)*dx+labelDx, nSig1);
    maxWidth = Math.max(maxWidth, g.getFontMetrics().stringWidth(str));

    // increasing labelDx, if labels would overlap 
    while (maxWidth > 0.7 * labelDx/(xmax-xmin)*width)
      {
      labelBase++;
      labelDx = Math.pow(10, intDx) * base[labelBase];
      if (labelBase % 3 == 2 && majorBase % 3 == 1)
        {
        majorBase++;
        majorDx = Math.pow(10, intDx) * base[majorBase];
        }
      }

    double xAct = Math.floor(xmin/dx)*dx;

    // draw base line
    g.drawLine(x1, y1, x1+width, y1);

    do {
      xScreen = (int) ((xAct-xmin)/(xmax-xmin)*width + x1 + 0.5);

      if (xScreen > x1 + width + 0.001) break;

      if (xScreen >= x1) {
        
        if (Math.abs(Math.floor(xAct/majorDx+0.5) - xAct/majorDx) <
            dx/majorDx/10.0) {

          if (Math.abs(Math.floor(xAct/labelDx+0.5) - xAct/labelDx) <
              dx/labelDx/10.0) {
            
            // grid lines
            if (tickGrid != 0 && xScreen > x1 && xScreen < x1 + width) {
              g.setColor(Color.lightGray);
              g.drawLine(xScreen, y1, xScreen, y1 - tickGrid);
              g.setColor(Color.black);
            }

            // label tick mark
            g.drawLine(xScreen, y1, xScreen, y1 - tickLabel);

            // label
            if (distLabel != 0) {
              str = expFormat(xAct, nSig1);
              
              strWidth = g.getFontMetrics().stringWidth(str);
              g.drawString(str, xScreen-strWidth/2, y1 - distLabel);
            }

          } else {
          
            // major tick mark
            g.drawLine(xScreen, y1, xScreen, y1 - tickMajor);

          }

        }
        else
          //  minor tick mark 
          g.drawLine(xScreen, y1, xScreen, y1 - tickMinor);
      }

      xAct += dx;

      // supress 1.23E-17 ... 
      if (Math.abs(xAct) < dx/100) 
        xAct=0;

    } while(true);
  }

  private int drawVAxis(Graphics g, int x1, int y1, int height, int tickGrid, 
                        double ymin, double ymax) {

    double labelDy, majorDy;
    int    yScreen, strWidth, maxWidth;
    String str;
    double base[] = {1,2,5,10,20,50,100,200,500,1000};

    int minTickDist = 5;
    int tickLabel   = 10;
    int tickMajor   = 10;
    int tickMinor   = 5;
    int distLabel   = -3;

    int ascent = g.getFontMetrics().getAscent();

    int maxTick = (int) ((double) (height/minTickDist) + 1);

    double dy = (ymax - ymin)/ (double) (height/minTickDist);

    double tmp = Math.log(dy) / LN10;
    int intDy = (int) tmp;
    double fracDy = tmp - intDy;
    
    if (fracDy <0) { 
      fracDy +=1; 
      intDy -=1; 
    }

    int tickBase = fracDy < LOG2 ? 1 :
                   fracDy < LOG5 ? 2 : 3;
    int majorBase = tickBase + 1;
    int labelBase = tickBase + 1;

    // rounding up of dy, labelDy 
    dy = Math.pow(10, intDy) * base[tickBase];
    majorDy = Math.pow(10, intDy) * base[majorBase];
    labelDy = majorDy;

    // number of significant digits 
    int nSig1 = 0;
    if (ymin > 0)
      nSig1 = (int) Math.floor(Math.log(Math.abs(ymin))/LN10) -
              (int) Math.floor(Math.log(Math.abs(labelDy))/LN10) + 1;

    int nSig2 = 0;
    if (ymax > 0)
      nSig2 = (int) Math.floor(Math.log(Math.abs(ymax))/LN10) -
              (int) Math.floor(Math.log(Math.abs(labelDy))/LN10) + 1;

    nSig1 = Math.max(nSig1, nSig2);
    nSig1 = Math.max(nSig1, 4);

    // determination of maximal width of labels 
    if (x1 == 0 && y1 == 0) {
      str = expFormat(Math.floor(ymin/dy)*dy, nSig1);
      maxWidth = g.getFontMetrics().stringWidth(str);

      str = expFormat(Math.floor(ymax/dy)*dy, nSig1);
      maxWidth = Math.max(maxWidth, g.getFontMetrics().stringWidth(str));

      str = expFormat(Math.floor(ymin/dy)*dy+labelDy, nSig1);
      maxWidth = Math.max(maxWidth, g.getFontMetrics().stringWidth(str));
      
      return maxWidth + Math.abs(distLabel);
    }

    int textHeight = g.getFontMetrics().getHeight();

    // increasing labelDy, if labels would overlap 
    while (textHeight*1.5 > labelDy/(ymax-ymin)*height)
      {
      labelBase++;
      labelDy = Math.pow(10, intDy) * base[labelBase];
      if (labelBase % 3 == 2 && majorBase % 3 == 1)
        {
        majorBase++;
        majorDy = Math.pow(10, intDy) * base[majorBase];
        }
      }

    double yAct = Math.floor(ymin/dy)*dy;

    // draw base line
    g.drawLine(x1, y1, x1, y1-height);

    do {
      yScreen = (int) (y1 - (yAct-ymin)/(ymax-ymin)*height + 0.5);

      if (yScreen < y1 - height - 0.001) break;

      if (yScreen <= y1) {
        
        if (Math.abs(Math.floor(yAct/majorDy+0.5) - yAct/majorDy) <
            dy/majorDy/10.0) {

          if (Math.abs(Math.floor(yAct/labelDy+0.5) - yAct/labelDy) <
              dy/labelDy/10.0) {
            
            // grid lines
            if (tickGrid != 0 && yScreen < y1 && yScreen > y1 - height) {
              g.setColor(Color.lightGray);
              g.drawLine(x1, yScreen, x1 + tickGrid, yScreen);
              g.setColor(Color.black);
            }

            // label tick mark
            g.drawLine(x1, yScreen, x1 + tickLabel, yScreen);

            // label
            if (distLabel != 0) {
              str = expFormat(yAct, nSig1);
              
              strWidth = g.getFontMetrics().stringWidth(str);
              if (distLabel < 0)
                g.drawString(str, x1-strWidth+distLabel, yScreen+ascent/2);
              else
                g.drawString(str, x1+distLabel, yScreen+ascent/2);
            }

          } else {
          
            // major tick mark
            g.drawLine(x1, yScreen, x1 + tickMajor, yScreen);
          }
        }
        else
          //  minor tick mark 
          g.drawLine(x1, yScreen, x1 + tickMinor, yScreen);
      }

      yAct += dy;

      // supress 1.23E-17 ... 
      if (Math.abs(yAct) < dy/100) 
        yAct=0;

    } while(true);
    
    return 0;
  }

  private String fixedFormat(double d, int precision)
  {  
    if (d > 0x7FFFFFFFFFFFFFFFL) 
      return expFormat(d, precision);
    if (precision == 0) 
      return (long)(d + 0.5) + "";

    long whole = (long)d;
    double fr = d - whole; // fractional part
    if (fr >= 1 || fr < 0) 
      return expFormat(d, precision);

    double factor = 1;
    String leading_zeroes = "";
    for (int i = 1; i <= precision && factor <= 0x7FFFFFFFFFFFFFFFL; i++) {  
      factor *= 10; 
      leading_zeroes = leading_zeroes + "0"; 
    }
    long l = (long) (factor * fr + 0.5);
    if (l >= factor) { 
      l = 0; 
      whole++; 
    } 
          
    String z = leading_zeroes + l;
    z = "." + z.substring(z.length() - precision, z.length());

    int t = z.length() - 1;
    while (t >= 0 && z.charAt(t) == '0') t--;
    if (t >= 0 && z.charAt(t) == '.') t--;
    z = z.substring(0, t + 1);

    return whole + z;
  }

  private String expFormat(double d, int precision)
    { 
    String s = "";
    String f = "";
    int e = 0;
    if (d < 0) {
      d = -d;
      s = "-";
    }
    double dd = d;
    double factor = 1;
    if (d != 0) {  
      while (dd > 10) { e++; factor /= 10; dd = dd / 10; }
      while (dd < 1) { e--; factor *= 10; dd = dd * 10; }
    }
    if (e >= -5 && e <= precision) 
      return s+fixedFormat(d, precision);
      
    d = d * factor;
    f = f + fixedFormat(d, precision);
     
    f = f + "E";

    String p = "00";      
    if (e >= 0) {  
      f = f + "+";
      p = p + e;
    } else {  
      f = f + "-";
      p = p + (-e);
    }
         
    return s + f + p.substring(p.length() - 2, p.length());
  }

  class DataPoint {
    public double x;
    public double y;
    
    DataPoint(double x, double y) {
      this.x = x;
      this.y = y;
    }

    DataPoint(String x, String y) {
      this.x = Double.valueOf(x).doubleValue();
      this.y = Double.valueOf(y).doubleValue();
    }
    
    public String toString() {
      return new String("["+x+","+y+"]");
    }
  }

  class DataCurve implements OdbKeyUpdate {
    
    private OdbKey odbKey;
    private String eventName;
    private String varName;
    private int    eventID;
    private Vector data;
    private int    interval;
    private Color  color;
    private double scaleYMin, scaleYMax;
    private MJHist mjhist;
    
    DataCurve(MJHist mjhist, String eventName, String varName, 
              Color color, int start, int end) 
        throws IOException, TooManyListenersException, OdbKeyNotFoundException {
      this.mjhist    = mjhist;
      this.eventName = eventName;
      this.varName   = varName;
      this.color     = color;
      this.scaleYMin = 0;
      this.scaleYMax = 0;
      this.data = new Vector();
      this.interval = (end-start)/500;

      // determine event ID
      String r = experiment.rpcCall("hs_get_event_id&0&"+eventName);
      this.eventID = Integer.parseInt(experiment.replyParam(r, 1));
        
      // create ODBKey and hot link
      try {
        if (eventName.equals("System"))
          odbKey = new OdbKey(experiment, "/History/Links");
        else
          odbKey = new OdbKey(experiment, "/Equipment/"+eventName+"/Variables");
        odbKey.hotLink(this);
      } catch (OdbKeyNotFoundException e) { 
        System.out.println("History event "+eventName+" not found in ODB, hotlink disabled"); 
        }
      
      getHistory(start, end, interval, varName);
    }

    public String getEventName() {
      return this.eventName;
    }

    public String getVarName() {
      return this.varName;
    }

    public Color getColor() {
      return this.color;
    }

    public void unlink() throws IOException {
      odbKey.hotUnlink();
    }
    
    private void getHistory(int start, int end, int interval, String name) throws IOException {

      String r = experiment.rpcCall("hs_read&"+eventID+"&"+start+"&"+end+"&"+interval+"&"+name+"&0&32000&32000");
    
      if (experiment.replyParam(r, 0).equals("1")) {

        int n = Integer.parseInt(experiment.replyParam(r, 1));
        double x[] = new double[n];
        double y[] = new double[n];

        int ib, ie;

        // skip status and n
        for (ib = 0 ; r.charAt(ib) != '&' ; ib++);
        for (ib++ ; r.charAt(ib) != '&' ; ib++);
        ib++;
        
        for (int i=0 ; i<n ; i++) {
          for (ie = ib+1; r.charAt(ie) != '&' ; ie++);
          x[i] = Double.valueOf(r.substring(ib, ie)).doubleValue();
          ib = ie+1;
        }
      
        // skip two parameters
        for (; r.charAt(ib) != '&' ; ib++);
        for (ib++ ; r.charAt(ib) != '&' ; ib++);
        ib++;
        
        for (int i=0 ; i<n ; i++) {
          for (ie = ib+1; r.charAt(ie) != '&' ; ie++);
          y[i] = Double.valueOf(r.substring(ib, ie)).doubleValue();
          ib = ie+1;
        }
        
        for (int i=0 ; i<n ; i++) {
          DataPoint dp = new DataPoint(x[i], y[i]);
          data.addElement(dp);

          if (dp.y < scaleYMin)
            scaleYMin = dp.y;
          if (dp.y > scaleYMax)
            scaleYMax = dp.y;
        }
      } else {
        if (experiment.replyParam(r, 0).equals("707"))
          System.out.println("Variable \""+name+"\" not found in history");
      }

    }
    
    public double getScaleYMin() {
      double min = 0;
      for (int i=0 ; i<data.size() ; i++) {
        DataPoint dp = (DataPoint) data.elementAt(i);
        if (i==0)
          min = dp.y;
        else if (dp.y < min)
          min = dp.y;
      }
      return min;
    }
    
    public double getScaleYMax() {
      double max = 0;
      for (int i=0 ; i<data.size() ; i++) {
        DataPoint dp = (DataPoint) data.elementAt(i);
        if (i==0)
          max = dp.y;
        else if (dp.y > max)
          max = dp.y;
      }
      return max;
    }

    public void paint(Graphics g, double xmin, double xmax, double ymin, double ymax,
                      int x, int y, int width, int height, int timeOffset) {

      g.setColor(this.color);
  
      int xd, yd, xold=0, yold=0, told=0, dt;
      int now = (int) (new Date().getTime()/1000) + timeOffset;
      DataPoint dp;
      
      for (int i=0 ; i<data.size() ; i++) {
      
        do {
          dp = (DataPoint) data.elementAt(i);
          
          xd = (int) (x + ((dp.x - now)/3600 - xmin)/(xmax - xmin)*width + 0.5);
          yd = (int) (y - (dp.y - ymin)/(ymax - ymin)*height + 0.5);
        
          // throw away old data points
          if (xd <= x)
            data.removeElementAt(i);

        } while (xd <= x);
      
        dt = (int)dp.x - told;

        // if more than 2min hole, don't draw line
        if (i==0 || dt>interval+80)
          g.drawLine(xd, yd, xd, yd);
        else
          g.drawLine(xold, yold, xd, yd);
      
        xold = xd;
        yold = yd;
        told = (int)dp.x;
      }
    } 
 
    public void update(OdbKey key)
    {

      try {
        int start;

        /* get new data points since last update */
        String r = experiment.rpcCall("cm_time");
        int end = Integer.parseInt(experiment.replyParam(r,1))+10;

        if (data.size() > 0) {
          DataPoint dp = (DataPoint)data.lastElement();
          start = (int) dp.x + 1;
        } else {
          start = end - 3600;
        }

        getHistory(start, end, 1, varName);
      
        mjhist.repaint();
      
      } catch (IOException e) { System.out.println(e); };
    }
  
  }
  
  // bean properties
  private String equipment;

  // member variables
  private Experiment experiment;
  private OdbKey odbKey;
  private Vector dataCurve;
  private int timeOffset; 
  private double scaleXMin, scaleXMax, scaleYMin, scaleYMax;
  private Timer timer;
  private boolean drawLabels;
  private boolean startYZero;
  private JPopupMenu popup;
  
  private Color colors[] = { Color.red, Color.blue, Color.green, Color.cyan, 
                             Color.darkGray, Color.magenta, Color.orange, Color.pink };
                             
}